#!/usr/bin/env python

"""
Confluence Wiki syntax parsing.

Copyright (C) 2012 Paul Boddie <paul@boddie.org.uk>

This software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this library; see the file LICENCE.txt
If not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

--------

The basic procedure is as follows:

 1. Wiki pages are first split up into regions.
 2. Then, within these regions, the text is split into blocks.
    1. First, lists are identified.
    2. Additionally, other block-like elements are identified.
 3. Each block is then parsed.
"""

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

from xmlread import Parser
import re
import sys
import operator
import htmlentitydefs

URL_SCHEMES = ("http", "https", "ftp", "mailto")

# Section extraction.

sections_regexp_str = r"(?<!{){(?P<type>[^-_*+{}\n:]+)(:[^}\n]+)?}.*?{(?P=type)}"
sections_regexp = re.compile(sections_regexp_str, re.DOTALL | re.MULTILINE)

def get_regions(s):

    """
    Return a list of regions from 's'. Each region is specified using a tuple of
    the form (type, text).
    """

    last = 0
    regions = []
    for match in sections_regexp.finditer(s):
        start, end = match.span()
        regions.append((None, s[last:start]))
        regions.append(get_section_details(s[start:end]))
        last = end
    regions.append((None, s[last:]))
    return regions

# Section inspection.

section_regexp_str = r"{(?P<sectiontype>[^\n:]*?)(?::(?P<options>.*?))?}(?P<section>.*){(?P=sectiontype)}"
section_regexp = re.compile(section_regexp_str, re.DOTALL | re.MULTILINE)

def get_section_details(s):

    "Return the details of a section 's' in the form (type, text)."

    match = section_regexp.match(s)
    if match:
        return (match.group("sectiontype"), match.group("options")), match.group("section")
    else:
        return None, s

# Heading, table and list extraction.

list_regexp_str = r"^\s*(?P<listtype>[*#-])[*#-]*.*\n(\s*(?P=listtype).*(?:\n|$))*"
table_regexp_str = r"^((?P<celltype>[|]{1,2})(.+?(?P=celltype))+(\n|$))+"
blocktext_regexp_str = r"^(?P<type>h\d|bq)\.\s+(?P<text>.*)$"

blockelement_regexp = re.compile(
    "(" + list_regexp_str + ")"
    "|"
    "(" + table_regexp_str + ")"
    "|"
    "(" + blocktext_regexp_str + ")",
    re.MULTILINE
    )

def get_block_elements(s):

    """
    Extract headings, tables and lists from the given string 's'.
    """

    last = 0
    blocks = []
    for match in blockelement_regexp.finditer(s):
        start, end = match.span()
        matchtype = match.group("listtype") and "list" or match.group("celltype") and "table" or match.group("type")
        blocks.append((None, s[last:start]))
        blocks.append((matchtype, match.group("text") or s[start:end]))
        last = end
    blocks.append((None, s[last:]))
    return blocks

# Block extraction.

block_regexp_str = r"^(?:\s*\n)+"
block_regexp = re.compile(block_regexp_str, re.MULTILINE)

def get_basic_blocks(s):

    """
    Return blocks from the given string 's' by splitting the text on blank lines
    and eliminating those lines.
    """

    return [b for b in block_regexp.split(s) if b.strip()]

# Block inspection.

def get_blocks(s):

    """
    Return blocks from the given string 's', inspecting the basic blocks and
    generating additional block-level text where appropriate.
    """

    blocks = []

    for blocktype, blocktext in get_block_elements(s):

        # Collect heading, list and table blocks.

        if blocktype is not None:
            blocks.append((blocktype, blocktext))

        # Attempt to find new subblocks in other regions.

        else:
            for block in get_basic_blocks(blocktext):
                blocks.append((None, block))

    return blocks

# List item inspection.

listitem_regexp_str = r"^(?P<marker> *[-*#]+)\s*(?P<text>.*)$"
listitem_regexp = re.compile(listitem_regexp_str, re.MULTILINE)

def get_list_items(text):

    "Return a list of (marker, text) tuples for the given list 'text'."

    items = []

    for match in listitem_regexp.finditer(text):
        items.append((match.group("marker"), match.group("text")))

    return items

# Table row inspection.

monospace_regexp_str = r"{{(?P<monotext>.*?)}}"
link_regexp_str = r"[[](?P<linktext>.*?)]"
image_regexp_str = r"!(?P<imagetext>.*?)!"
cellsep_regexp_str = r"(?P<celltype>[|]{1,2})"

content_regexp_str = (
    "(" + monospace_regexp_str + ")"
    "|"
    "(" + link_regexp_str + ")"
    "|"
    "(" + image_regexp_str + ")"
    )

table_content_regexp_str = (
    content_regexp_str +
    "|"
    "(" + cellsep_regexp_str + ")"
    )

content_regexp = re.compile(content_regexp_str)
table_content_regexp = re.compile(table_content_regexp_str)

def translate_content_match(match):

    "Translate the content described by the given 'match', returning a string."

    if match.group("monotext"):
        return "{{{%s}}}" % match.group("monotext")

    elif match.group("linktext"):
        parts = match.group("linktext").split("|")

        # NOTE: Proper detection of external links required.

        if len(parts) == 1:
            label, target, title = None, parts[0], None
        elif len(parts) == 2:
            (label, target), title = parts, None
        else:
            label, target, title = parts

        target = target.strip()

        # Look for namespace links and rewrite them.

        if target.find(":") != -1:
            prefix = ""
            space, rest = target.split(":", 1)
            if space not in URL_SCHEMES:
                target = "%s/%s" % (space, rest)

        # Detect anchors.

        elif target.startswith("#"):
            prefix = ""

        # Detect attachments.

        elif target.startswith("^"):
            prefix = "attachment:"

        # Link to other pages within a space.

        else:
            prefix = "../"

            # Make the link tidier by making a target if none was given.

            if not label:
                label = target

        if not label and not title:
            return "[[%s%s]]" % (prefix, target)
        elif not title:
            return "[[%s%s|%s]]" % (prefix, target, label)
        else:
            return "[[%s%s|%s|title=%s]]" % (prefix, target, label, title)

    elif match.group("imagetext"):
        parts = match.group("imagetext").split("|")

        # NOTE: Proper detection of external links required.

        if parts[0].startswith("http"):
            prefix = ""
        else:
            prefix = "attachment:"

        # NOTE: Proper options conversion required.

        if len(parts) == 1:
            return "{{%s%s}}" % (prefix, parts[0])
        else:
            return "{{%s%s|%s}}" % (prefix, parts[0], parts[1])

    else:
        return match.group()

def get_table_rows(text):

    "Return a list of (cellsep, columns) tuples for the given table 'text'."

    rows = []

    for line in text.split("\n"):
        cellsep = None
        columns = [""]
        last = 0
        for match in table_content_regexp.finditer(line):
            start, end = match.span()
            columns[-1] += line[last:start]

            if match.group("celltype"):
                if cellsep is None:
                    cellsep = match.group("celltype")
                columns.append("")
            else:
                columns[-1] += match.group()

            last = end

        columns[-1] += line[last:]

        if cellsep:
            rows.append((cellsep, columns[1:-1]))

    return rows

def translate_content(text, sectiontype=None):

    """
    Return a translation of the given 'text'. If the optional 'sectiontype' is
    specified, the translation may be modified to a form appropriate to the
    section being translated.
    """

    parts = []

    last = 0
    for match in content_regexp.finditer(text):
        start, end = match.span()
        parts.append(text[last:start])

        # Handle unformatted sections.

        if sectiontype in ("code", "noformat"):
            parts.append(match.group())
        else:
            parts.append(translate_content_match(match))

        last = end

    parts.append(text[last:])
    return "".join(parts)

# Translation helpers.

blocktypes = {
    "h1" : "= %s =",
    "h2" : "== %s ==",
    "h3" : "=== %s ===",
    "h4" : "==== %s ====",
    "h5" : "===== %s =====",
    "h6" : "====== %s ======",
    "bq" : "{{{%s}}}",
    }

markers = {
    "*" : "*",
    "#" : "1.",
    "-" : "*",
    }

def translate_marker(marker):

    "Translate the given 'marker' to a suitable Moin representation."

    return " " * len(marker) + markers[marker[-1]]

cellseps = {
    "|" : "||",
    "||" : "||",
    }

cellextra = {
    "|" : "",
    "||" : "'''",
    }

def translate_cellsep(cellsep):

    "Translate the given 'cellsep' to a suitable Moin representation."

    return cellseps[cellsep]

def translate_cell(cellsep, text):

    "Using 'cellsep', translate the cell 'text'."

    return cellextra[cellsep] + translate_content(text) + cellextra[cellsep]

sectiontypes = {
    "code" : "",
    "noformat" : "",
    "quote" : "",
    "info" : "wiki important",
    "note" : "wiki caution",
    "tip" : "wiki tip",
    "warning" : "wiki warning",
    }

# XML dialect syntax parsing.

tags = {
    # XHTML tag               MoinMoin syntax
    "strong"                : "'''%s'''",
    "em"                    : "''%s''",
    "u"                     : "__%s__",
    "del"                   : "--(%s)--",
    "sup"                   : "^%s^",
    "sub"                   : ",,%s,,",
    "code"                  : "`%s`",
    "pre"                   : "{{{%s}}}",
    "blockquote"            : " %s",
    "small"                 : "~-%s-~",
    "big"                   : "~+%s+~",
    "p"                     : "\n%s\n",
    "ol"                    : "\n%s",
    "ul"                    : "\n%s",
    "ac:plain-text-body"    : "{{{%s}}}",
    "ac:link"               : "[[%s%s|%s]]",
    }

for tag, translation in blocktypes.items():
    tags[tag] = "\n%s\n" % translation

simple_tags = {
    # XHTML tag               MoinMoin syntax
    "br"                    : "<<BR>>",
    }

list_tags = {
    # XHTML list tag          MoinMoin list item syntax
    "ol"                    : "1. %s\n",
    "ul"                    : "* %s\n",
    }

indented_tags = ["li", "p"]

link_target_tags = {
    # Confluence element      Attribute providing the target
    "ri:page"               : "ri:content-title",
    "ri:attachment"         : "ri:filename",
    "ri:user"               : "ri:username",
    }

macro_rich_text_styles = {
    # Confluence style        MoinMoin admonition style
    "note"                  : "caution",
    "warning"               : "warning",
    "info"                  : "important",
    "tip"                   : "tip",
    }

normalise_regexp_str = r"\s+"
normalise_regexp = re.compile(normalise_regexp_str)

normalise_end_regexp_str = r"\s\s+$"
normalise_end_regexp = re.compile(normalise_end_regexp_str)

class ConfluenceXMLParser(Parser):

    "Handle content from Confluence 4 page revisions."

    def __init__(self, out):
        Parser.__init__(self)
        self.out = out

        # Link target information.

        self.target = None
        self.target_type = None

        # Macro information.

        self.macro = None
        self.macro_parameters = {}

        # Indentation and preformatted states.

        self.indent = 0
        self.states = {}
        for name in ("pre", "ac:plain-text-body"):
            self.states[name] = 0

    # ContentHandler-related methods.

    def startElement(self, name, attrs):
        if list_tags.has_key(name):
            self.indent += 1
        elif self.states.has_key(name):
            self.states[name] += 1
        Parser.startElement(self, name, attrs)

    def endElement(self, name):
        Parser.endElement(self, name)
        if list_tags.has_key(name):
            self.indent -= 1
        elif self.states.has_key(name):
            self.states[name] -= 1

    def characters(self, content):
        if not self.is_preformatted():
            content = self.normalise(content, self.elements[-1])
        Parser.characters(self, content)

    def skippedEntity(self, name):
        ch = htmlentitydefs.name2codepoint.get(name)
        if ch:
            self.text[-1].append(unichr(ch))

    # Parser-related methods.

    def handleElement(self, name):
        text = "".join(self.text[-1])
        conversion = None

        # Handle list elements.

        if name == "li" and len(self.elements) > 1:
            list_tag = self.elements[-2]
            conversion = list_tags.get(list_tag)

        # Remember link target information.

        elif link_target_tags.has_key(name):
            self.target = self.attributes[-1].get(link_target_tags[name])
            self.target_type = name
            text = ""

        # Remember macro information.

        elif name == "ac:parameter":
            self.macro_parameters[self.attributes[-1].get("ac:name")] = text
            text = ""

        elif name == "ac:macro":
            self.macro = self.attributes[-1].get("ac:name")

        # Handle the common case.

        else:
            conversion = tags.get(name)

        # Attempt to convert the text.

        # Links require target information.
        # NOTE: User links should support the intended user namespace prefix.

        if name == "ac:link":
            if self.target_type == "ri:attachment":
                prefix = "attachment:"
            elif self.target_type == "ri:user":
                prefix = ""
            else:
                prefix = "../"

            text = conversion % (prefix, self.target, text or self.target)
            self.target = self.target_type = None

        # Macro name information is used to style rich text body regions.

        elif name == "ac:macro" and macro_rich_text_styles.has_key(self.macro):
            details = macro_rich_text_styles[self.macro]
            title = self.macro_parameters.get("title")
            if title:
                details = "%s\n\n%s" % (details, title)
            text = "{{{#!wiki %s\n\n%s}}}" % (details, text)
            self.macro = None
            self.macro_parameters = {}

        # Handle the common case.

        elif text and conversion:
            text = conversion % text
        elif simple_tags.has_key(name):
            text = simple_tags[name]

        # Normalise leading whitespace and indent the text if appropriate.

        if name in indented_tags:
            text = " " * self.indent + text.lstrip()

        # Add the converted text to the end of the parent element's text nodes.

        if len(self.text) > 1:
            preceding = "".join(self.text[-2])

            if not self.is_preformatted():
                preceding = self.normalise_end(preceding, self.elements[-2])

            self.text[-2] = [preceding]
            self.text[-2].append(text)

        # Otherwise, emit the text.

        else:
            self.out.write(text)

    def is_preformatted(self):
        return reduce(operator.or_, self.states.values(), False)

    def get_replacement(self, name, end=False):
        if list_tags.has_key(name):
            if end:
                return "\n"
            else:
                return ""
        elif name == "body":
            return "\n\n"
        else:
            return " "

    def normalise(self, text, name):
        return normalise_regexp.sub(self.get_replacement(name), text)

    def normalise_end(self, text, name):
        return normalise_end_regexp.sub(self.get_replacement(name, True), text)

def xmlparse(s, out):

    "Parse the content in the string 's', writing a translation to 'out'."

    # NOTE: CDATA sections appear to have erroneous endings.

    s = u"""\
<?xml version="1.0"?>
<!DOCTYPE html 
     PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"
     "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
<body>
%s
</body>
</html>""" % s.replace("]] >", "]]>")

    f = StringIO(s.encode("utf-8"))
    try:
        parser = ConfluenceXMLParser(out)
        parser.parse(f)
    finally:
        f.close()

# General parsing.

def parse(s, out):

    "Parse the content in the string 's', writing a translation to 'out'."

    for type, text in get_regions(s):

        # Handle list, heading, blockquote or anonymous blocks.

        if type is None:
            for blocktype, blocktext in get_blocks(text):

                # Translate headings and blockquotes.

                if blocktypes.has_key(blocktype):
                    print >>out, blocktypes[blocktype] % blocktext

                # Translate list items.

                elif blocktype == "list":
                    for listmarker, listitem in get_list_items(blocktext):
                        print >>out, "%s %s" % (translate_marker(listmarker), translate_content(listitem))

                # Translate table items.

                elif blocktype == "table":
                    for cellsep, columns in get_table_rows(blocktext):
                        moinsep = translate_cellsep(cellsep)
                        print >>out, moinsep + moinsep.join([translate_cell(cellsep, column) for column in columns]) + moinsep

                # Handle anonymous blocks.

                else:
                    print >>out, translate_content(blocktext.rstrip())

                print >>out

        # Handle sections.

        else:
            sectiontype, options = type

            # Direct translations of sections.

            mointype = sectiontypes.get(sectiontype)
            if mointype:
                print >>out, "{{{#!%s" % mointype
                if options:
                    print >>out, "##", options
            else:
                print >>out, "{{{",
            print >>out, translate_content(text, sectiontype),
            print >>out, "}}}"
            print >>out

if __name__ == "__main__":
    s = sys.stdin.read()
    parse(s, sys.stdout)

# vim: tabstop=4 expandtab shiftwidth=4
